package speiger.src.coreengine.rendering.textures.custom;

import java.nio.ByteBuffer;
import java.util.Iterator;

import org.lwjgl.opengl.GL11;
import org.lwjgl.system.MemoryStack;
import org.lwjgl.system.MemoryUtil;

import speiger.src.collections.ints.collections.IntIterator;
import speiger.src.collections.ints.maps.impl.hash.Int2ObjectOpenHashMap;
import speiger.src.collections.ints.maps.interfaces.Int2ObjectMap;
import speiger.src.collections.ints.maps.interfaces.Int2ObjectMap.Entry;
import speiger.src.collections.ints.sets.IntOpenHashSet;
import speiger.src.collections.ints.sets.IntSet;
import speiger.src.collections.ints.utils.maps.Int2ObjectMaps;
import speiger.src.collections.utils.ITrimmable;
import speiger.src.coreengine.math.BitUtil;
import speiger.src.coreengine.rendering.textures.base.AbstractTexture;
import speiger.src.coreengine.rendering.textures.base.TextureManager;
import speiger.src.coreengine.rendering.utils.AllocationTracker;

public class DynamicTexture extends AbstractTexture
{
	Int2ObjectMap<IntSet> dirtyChunks = new Int2ObjectOpenHashMap<IntSet>();
	int[] data;
	int width;
	int height;
	boolean first = true;
	
	public DynamicTexture(int width, int height)
	{
		this.width = width;
		this.height = height;
		setTextureID(GL11.glGenTextures());
		data = new int[width * height];
	}
	
	public int[] getTextureData()
	{
		return data;
	}
	
	public void markDirty(int x, int z)
	{
		int chunkIndex = BitUtil.toInt(x >> 4, z >> 4);
		IntSet set = dirtyChunks.get(chunkIndex);
		if(set == null)
		{
			set = new IntOpenHashSet();
			dirtyChunks.put(chunkIndex, set);
		}
		set.add(BitUtil.toInt(x, z));
	}
	
	public void setData(int x, int z, int value)
	{
		data[(z * width) + x] = value;
		markDirty(x, z);
	}
	
	public boolean hasTasks()
	{
		return dirtyChunks.size() > 0 || first;
	}
	
	public void updateData()
	{
		if(first)
		{
			ByteBuffer buffer = MemoryUtil.memAlloc(data.length * 4);
			for(int i = 0;i<data.length;i++)
			{
				int pixel = data[i];
				buffer.put((byte)((pixel >> 16) & 0xFF));
				buffer.put((byte)((pixel >> 8) & 0xFF));
				buffer.put((byte)(pixel & 0xFF));
				buffer.put((byte)((pixel >> 24) & 0xFF));
			}
			buffer.flip();
			bindTexture();
			GL11.glTexParameterf(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_NEAREST);
			GL11.glTexParameterf(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_NEAREST);
			AllocationTracker.INSTANCE.addGPUBytes(buffer.remaining());
			GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, GL11.GL_RGBA, width, height, 0, GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, buffer);
			first = false;
			MemoryUtil.memFree(buffer);
			dirtyChunks.clear();
		}
		else
		{
			bindTexture();
			Thread thread = Thread.currentThread();
			for(Iterator<Entry<IntSet>> iter = Int2ObjectMaps.fastIterator(dirtyChunks);iter.hasNext() && !thread.isInterrupted();iter.remove())
			{
				Entry<IntSet> values = iter.next();
				int key = values.getIntKey();
				int chunkX = BitUtil.toFirstShort(key) * 16;
				int chunkZ = BitUtil.toSecondShort(key) * 16;
				int width = 0;
				int height = 0;
				for(IntIterator subIt = values.getValue().iterator();subIt.hasNext();)
				{
					int index = subIt.nextInt();
					width = Math.max(width, BitUtil.toFirstShort(index) - chunkX + 1);
					height = Math.max(height, BitUtil.toSecondShort(index) - chunkZ + 1);
				}
				try(MemoryStack stack = MemoryStack.stackPush())
				{
					ByteBuffer buffer = stack.malloc(width * height * 4);
					for(int y = 0;y<height;y++)
					{
						for(int x = 0;x<width;x++)
						{
							int pixel = data[((chunkZ + y) * this.width) + chunkX + x];
							buffer.put((byte)((pixel >> 16) & 0xFF));
							buffer.put((byte)((pixel >> 8) & 0xFF));
							buffer.put((byte)(pixel & 0xFF));
							buffer.put((byte)((pixel >> 24) & 0xFF));
						}
					}
					buffer.flip();
					AllocationTracker.INSTANCE.addGPUBytes(buffer.remaining());
					GL11.glTexSubImage2D(GL11.GL_TEXTURE_2D, 0, chunkX, chunkZ, width, height, GL11.GL_RGBA, GL11.GL_UNSIGNED_BYTE, buffer);
				}
			}
		}
		if(dirtyChunks.isEmpty())
		{
			((ITrimmable)dirtyChunks).trim();
		}
	}
	
	@Override
	public void reload()
	{
		int old = getTextureID();
		setTextureID(GL11.glGenTextures());
		TextureManager.INSTANCE.removeTexture(old);
		dirtyChunks.clear();
		first = true;
		updateData();
	}
	
	@Override
	public int getWidth()
	{
		return width;
	}

	@Override
	public int getHeight()
	{
		return height;
	}
}
